# -*- coding: utf-8 -*-

"""
    FanFilm Add-on

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import base64
import hashlib
import json
import re

import xbmc

import requests
from urllib.parse import urlparse, quote_plus
import xbmcaddon
from ptw.debug import log_exception
from ptw.libraries import client
from ptw.libraries import directstream
from ptw.libraries import pyaes
from ptw.libraries import trakt
from ptw.libraries import cleantitle
from ptw.libraries import control

RES_4K = [
    " 4k",
    " hd4k",
    " 4khd",
    " uhd",
    " ultrahd",
    " ultra hd",
    " 2160",
    " 2160p",
    " hd2160",
    " 2160hd",
]
RES_1080 = [
    " 1080",
    " 1080p",
    " 1080i",
    " hd1080",
    " 1080hd",
    " m1080p",
    " fullhd",
    " full hd",
    " fhd",
    " 1o8o",
    " 1o8op",
]
RES_720 = [" 720", " 720p", " 720i", " hd720", " 720hd", " 72o", " 72op"]
RES_SD = [
    " 576",
    " 576p",
    " 576i",
    " sd576",
    " 576sd",
    " 480",
    " 480p",
    " 480i",
    " sd480",
    " 480sd",
    " 360",
    " 360p",
    " 360i",
    " sd360",
    " 360sd",
    " 240",
    " 240p",
    " 240i",
    " sd240",
    " 240sd",
]
SCR = [" scr", " screener", " dvdscr", " dvd scr", " r5", " r6"]
CAM = [
    " camrip",
    " tsrip",
    " hdcam",
    " hqcam",
    " hd cam",
    " cam rip",
    " hdts",
    " dvdcam",
    " dvdts",
    " cam",
    " telesync",
    " ts",
]
AVC = [" h 264 ", " h264 ", " x264 ", " avc "]


def supported_video_extensions():
    supported_video_extensions = xbmc.getSupportedMedia("video").split("|")
    unsupported = [
        "",
        ".url",
        ".bin",
        ".zip",
        ".rar",
        ".001",
        ".disc",
        ".7z",
        ".tar.gz",
        ".tar.bz2",
        ".tar.xz",
        ".tgz",
        ".tbz2",
        ".gz",
        ".bz2",
        ".xz",
        ".tar",
    ]
    return [i for i in supported_video_extensions if i not in unsupported]


def is_anime(content, type, type_id):
    try:
        r = trakt.getGenre(content, type, type_id)
        return "anime" in r or "animation" in r
    except Exception:
        log_exception()
        return False


def absoluteNumber(tvdb, episode, season):
    headers = {
        "Content-Type": "application/json",
        "Accept": "application/json",
    }
    ADDON = xbmcaddon.Addon("script.module.ptw")
    apikey = str(ADDON.getSetting("ApiKey"))
    userkey = str(ADDON.getSetting("UserKey"))
    username = str(ADDON.getSetting("Username"))
    data = '{\n  "apikey": "%s",\n  "userkey": "%s",\n  "username": "%s"\n}' % (
        apikey,
        userkey,
        username,
    )
    tvdbapi = json.loads(
        requests.post(
            "https://api.thetvdb.com/login", headers=headers, data=data
        ).content
    )["token"]
    headers = {"Authorization": "Bearer " + tvdbapi, "Accept": "application/json"}
    test2 = json.loads(
        requests.get(
            "https://api.thetvdb.com/series/%s/episodes/query?airedSeason=%s&airedEpisode=%s"
            % (tvdb, season, episode),
            headers=headers,
        ).content
    )
    odcinek = str(test2["data"][0]["absoluteNumber"])
    return odcinek


def get_qual(term):
    term = " {} ".format(term).lower()
    if any(i in term for i in SCR):
        return "SCR"
    elif any(i in term for i in CAM):
        return "CAM"
    elif any(i in term for i in RES_4K) and not any(i in term for i in RES_1080):
        return "4K"
    elif any(i in term for i in RES_1080):
        return "1080p"
    elif any(i in term for i in RES_720):
        return "720p"
    elif any(i in term for i in RES_SD):
        return "SD"
    elif "remux " in term and any(i in term for i in AVC):
        return "1080p"
    elif "remux " in term:
        return "4K"
    else:
        return "SD"


def get_release_quality(release_name, release_link=""):
    if not release_name and not release_link:
        return "SD", []

    try:
        if release_link:
            term = " ".join(
                (cleantitle.get_title(release_name), cleantitle.get_title(release_link))
            )
        else:
            term = cleantitle.get_title(release_name)

        quality = get_qual(term)
        if not quality:
            quality = "SD"

        info = []
        if "3d" in term or ".3D." in release_name:
            info.append("3D")
        if any(i in term for i in ["hevc", "h265", "x265"]):
            info.append("HEVC")

        return quality, info
    except Exception:
        return "SD", []


def getFileType(url):
    try:
        url = url.lower()
        url = url.replace(" ", ".")  # for easier detection
        url = url.replace("_", ".")  # the same as above
        url = url.replace(".1080", ".Full")  # for case ..1995.1080p.. for 5.1 channels
        url = url.replace(".1440", "..1440")  # for case ..1995.1440p.. for 5.1 channels
        url = url.replace(".10bit", "..10bit")  # for case ..265.10bit..  for 5.1 channels
        url = url.replace(".12bit", "..12bit")  # for case ..265.12bit..  for 5.1 channels
        url = url.replace(".x26", "..x26")  # for case dts.x264
        ext = re.search(r"\.\w{2,4}$", url)  # search extension
        ext = ext[0] if ext else ''  # remember extension
        url = re.sub(r"\.\w{2,4}$", ".", url)  # extension out
    except Exception:
        url = str(url)
    type = ""

    if "bluray" in url:
        type += " BLURAY /"
    if "blu-ray" in url:
        type += " BLU-RAY /"
    if ".web-dl" in url:
        type += " WEB-DL /"
    if ".web." in url:
        type += " WEB /"
    if "dvdrip" in url:
        type += " DVDRip /"
    if "dvd-rip" in url:
        type += " DVD-Rip /"
    if "hdrip" in url:
        type += " HDRip /"
    if "bd-r." in url:
        type += " BD-R /"
    if "bd-rip" in url:
        type += " BD-RIP /"
    if "bd.r." in url:
        type += " BD-R /"
    if "bd.rip" in url:
        type += " BD-RIP /"
    if "bdr." in url:
        type += " BD-R /"
    if "bdrip" in url:
        type += " BDRIP /"
    if any(i in url for i in ["dcprip", "dcp-rip", "dcp.rip"]):
        type += " DCP-Rip /"
    if "hdtv" in url:
        type += " HDTV /"
    if "tvrip" in url:
        type += " TVRip /"
    if any(i in url for i in ["camrip", "tsrip", "hdcam", "hqcam", "hdts", "dvdcam", "dvdts", "cam",
                              "telesync", ".ts"]):
        type += " cam /"
    if any(i in url for i in [".scr.", "scr.", ".screener", "dvdscr", ".r5.", ".r6."]):
        type += " SCR /"
    if ".md" in url:
        type += " MD /"
    if any(i in url for i in ["custom audio", "custom.audio", "custom-audio", "custom_audio"]):
        type += " custom audio /"
    if any(i in url for i in ["ac3", "ac-3"]):
        if any(i in url for i in ["eac3", "e-ac3", "eac-3", "e-ac-3"]):
            type += " E-AC3 /"
        else:
            type += " AC3 /"
    if ".dd" in url:
        if ".ddp" in url:
            type += " DD+ /"
        else:
            if any(i in url for i in ["dd-ex", "dd.ex"]):
                type += " DD-EX /"
            else:
                type += " DD /"
    if ".dts" in url:
        if any(i in url for i in ["dts-hd", "dts.hd."]):
            if any(i in url for i in ["dts-hd.ma", "dts.hd.ma"]):
                type += " DTS-HD MA /"
            else:
                type += " DTS-HD /"
        elif any(i in url for i in ["dts-es", "dts.es"]):
            type += " DTS-ES /"
        elif any(i in url for i in ["dts-ex", "dts.ex"]):
            type += " DTS-EX /"
        elif any(i in url for i in ["dts-x", "dts.x", "dtsx"]):
            type += " DTS-X /"
        else:
            type += " DTS /"
    if ".truehd" in url:
        type += " TrueHD /"
    if ".lpcm" in url:
        type += " LPCM /"
    if ".aac" in url:
        type += " AAC /"
    if "5.1" in url:
        type += " 5.1 /"
    if "6.1" in url:
        type += " 6.1 /"
    if "7.1" in url:
        type += " 7.1 /"
    if "2.0" in url:
        type += " 2.0 /"
    if "5.0" in url:
        type += " 5.0 /"
    if "1.0" in url:
        type += " 1.0 /"
    if "atmos" in url:
        type += " ATMOS /"
    if "hdr" in url:
        type += " HDR /"
    if ".dv." in url or "dolbyvision" in url or ".dovi." in url:
        type += " DV /"
    if ".imax." in url:
        type += " IMAX /"
    if ".3d." in url:
        type += " 3D /"
    if "sub" in url:
        if type != "":
            type = type.rstrip("/")
            type += "| SUBS |"
        else:
            type = " SUBS |"
#    if ".multi." in url:
#        type += " MULTI /"
#    if ".dual." in url:
#        type += " DUAL /"
    if ".xvid" in url:
        type += " XVID /"
    if ".h.264" in url:
        type += " H.264 /"
    if ".h.265" in url:
        type += " H.265 /"
    if ".h264" in url:
        type += " H264 /"
    if ".x264" in url:
        type += " x264 /"
    if ".x265" in url:
        type += " x265 /"
    if ".h265" in url:
        type += " h265 /"
    if ".hevc" in url:
        type += " hevc /"
    if ".avc." in url:
        type += " AVC /"
    if ".av1." in url:
        type += " AV1 /"
    if ".vc1." in url or ".vc-1." in url:
        type += " VC1 /"
    if any(i in url for i in ["mpeg2", "mpeg-2"]):
        type += " MPEG-2 /"
    if any(i in url for i in ["mpeg4", "mpeg-4"]):
        type += " MPEG-4 /"
    if any(i in url for i in ["mpeg1", "mpeg-1"]):
        type += " MPEG-1 /"
    if "48fps" in url:
        type += " 48 fps /"
    if "50fps" in url:
        type += " 50 fps /"
    if "60fps" in url:
        type += " 60 fps /"
    if "100fps" in url:
        type += " 100 fps /"
    if "120fps" in url:
        type += " 120 fps /"
    if "extended" in url:
        type += " extended /"
    if "directors.cut" in url:
        type += " directors cut /"
    if "remastered" in url:
        type += " REMASTERED /"
    if "rekonstrukcja" in url:
        type += " Rekonstrukcja cyfrowa /"
    if "remux" in url:
        type += " REMUX /"
    # if ".mp4" in ext:
        # type += " MP4 /"
    if ".avi" in ext:
        type += " AVI /"
    if ".ts" in ext:
        type += " .TS /"
    if "mpg" in ext:
        type += " MPG /"
    type = type.rstrip("/")
    return type


def check_sd_url(release_link):
    try:
        release_link = release_link.lower()
        release_link = re.sub(r"\.\w{2,4}$", "", release_link)  # extension out
        if "2160" in release_link:
            return "4K"
        elif "1440" in release_link:
            return "1440p"
        elif "1080i" in release_link:
            quality = "1080i"
        elif "1080" in release_link:
            quality = "1080p"
        elif "720" in release_link:
            quality = "720p"
        elif ".hd." in release_link:
            quality = "720p"
        elif any(i in release_link for i in ["dvdscr", "r5", "r6"]):
            quality = "SCR"
        elif any(
                i in release_link
                for i
                in [
                    "camrip",
                    "tsrip",
                    "hdcam",
                    "hqcam",
                    "hdts",
                    "dvdcam",
                    "dvdts",
                    "cam",
                    "telesync",
                    ".ts"
                ]
        ):
            quality = "CAM"
        else:
            quality = "SD"
        return quality
    except Exception:
        return "SD"


def label_to_quality(label):
    try:
        try:
            label = int(re.search(r"(\d+)", label).group(1))
        except Exception:
            label = 0

        if label >= 2160:
            return "4K"
        elif label >= 1440:
            return "1440p"
        elif label >= 1080:
            return "1080p"
        elif 720 <= label < 1080:
            return "720p"
        elif label < 720:
            return "SD"
    except Exception:
        return "SD"


def strip_domain(url):
    try:
        if url.lower().startswith("http") or url.startswith("/"):
            url = re.findall("(?://.+?|)(/.+)", url)[0]
        url = client.replaceHTMLCodes(url)
        url = url.encode("utf-8")
        return url
    except Exception:
        log_exception()


def is_host_valid(url, domains):
    try:
        host = __top_domain(url)
        hosts = [
            domain.lower() for domain in domains if host and host in domain.lower()
        ]

        if hosts and "." not in host:
            host = hosts[0]
        if hosts and any([h for h in ["google", "picasa", "blogspot"] if h in host]):
            host = "gvideo"
        if hosts and any([h for h in ["akamaized", "ocloud"] if h in host]):
            host = "CDN"
        return any(hosts), host
    except Exception:
        log_exception()
        return False, ""


def __top_domain(url):
    elements = urlparse(url)
    domain = elements.netloc or elements.path
    domain = domain.split("@")[-1].split(":")[0]
    regex = r"(?:www\.)?([\w\-]*\.[\w\-]{2,3}(?:\.[\w\-]{2,3})?)$"
    res = re.search(regex, domain)
    if res:
        domain = res.group(1)
    domain = domain.lower()
    return domain


def aliases_to_array(aliases, filter=None):
    try:
        if not filter:
            filter = []
        if isinstance(filter, str):
            filter = [filter]

        return [
            x.get("title") for x in aliases if not filter or x.get("country") in filter
        ]
    except Exception:
        log_exception()
        return []


def append_headers(headers):
    return "|%s" % "&".join(
        ["%s=%s" % (key, quote_plus(headers[key])) for key in headers]
    )


def get_size(url):
    try:
        size = client.request(url, output="file_size")
        if size == "0":
            size = False
        size = convert_size(size)
        return size
    except Exception:
        log_exception()
        return False


def convert_size(size_bytes):
    import math

    if size_bytes == 0:
        return "0B"
    size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
    i = int(math.floor(math.log(size_bytes, 1024)))
    p = math.pow(1024, i)
    s = round(size_bytes / p, 2)
    # if size_name[i] == "B" or size_name[i] == "KB":
    #    return None  # dlaczego?
    return "%s %s" % (s, size_name[i])


def check_directstreams(url, hoster="", quality="SD"):
    urls = []
    host = hoster

    if "google" in url or any(x in url for x in ["youtube.", "docid="]):
        urls = directstream.google(url)
        if not urls:
            tag = directstream.googletag(url)
            if tag:
                urls = [{"quality": tag[0]["quality"], "url": url}]
        if urls:
            host = "gvideo"
    elif "ok.ru" in url:
        urls = directstream.odnoklassniki(url)
        if urls:
            host = "vk"
    elif "vk.com" in url:
        urls = directstream.vk(url)
        if urls:
            host = "vk"
    elif any(x in url for x in ["akamaized", "blogspot", "ocloud.stream"]):
        urls = [{"url": url}]
        if urls:
            host = "CDN"

    direct = True if urls else False

    if not urls:
        urls = [{"quality": quality, "url": url}]

    return urls, host, direct


# if salt is provided, it should be string
# ciphertext is base64 and passphrase is string
def evp_decode(cipher_text, passphrase, salt=None):
    cipher_text = base64.b64decode(cipher_text)
    if not salt:
        salt = cipher_text[8:16]
        cipher_text = cipher_text[16:]
    data = evpKDF(passphrase, salt)
    decrypter = pyaes.Decrypter(pyaes.AESModeOfOperationCBC(data["key"], data["iv"]))
    plain_text = decrypter.feed(cipher_text)
    plain_text += decrypter.feed()
    return plain_text


def evpKDF(passwd, salt, key_size=8, iv_size=4, iterations=1, hash_algorithm="md5"):
    target_key_size = key_size + iv_size
    derived_bytes = ""
    number_of_derived_words = 0
    block = None
    hasher = hashlib.new(hash_algorithm)
    while number_of_derived_words < target_key_size:
        if block is not None:
            hasher.update(block)

        hasher.update(passwd)
        hasher.update(salt)
        block = hasher.digest()
        hasher = hashlib.new(hash_algorithm)

        for _i in range(1, iterations):
            hasher.update(block)
            block = hasher.digest()
            hasher = hashlib.new(hash_algorithm)

        derived_bytes += block[0 : min(len(block), (target_key_size - number_of_derived_words) * 4)]  # noqa: E203

        number_of_derived_words += len(block) / 4

    return {"key": derived_bytes[0 : key_size * 4], "iv": derived_bytes[key_size * 4 :]}  # noqa: E203


def get_lang_by_type(text):
    text = text.lower()
    if "dubbing" in text or "pldub" in text or "pl.dub" in text:
        if any(x in text for x in ["kino" ,"md"]):
            return "pl", "Dubbing Kino"
        return "pl", "Dubbing"
    elif "lektor pl" in text or "lektor" in text:
        return "pl", "Lektor"
    elif any(x in text for x in [".sub" ,"plsub", "subbed", "napisy pl", "napisy", "nap.pl"]):
        return "pl", "Napisy"
    elif any(x in text for x in ["polski", "polish", ".pl.", "pl.", "pl-", "pl_", "pl ", " pl "]):
        return "pl", ""
    # elif ".dual." in text:  # przenioslem wykrywanie do patterns.py
        # return "en", "dual"  # bo nie ma pewnosci, że jest pl
    elif any(x in text for x in [".multi." ,"_multi_", " multi ", ".mul.", " mul ", "_mul_"]):
        #return "en", "multi"  # bo nie ma pewnosci, że jest pl
        return "multi", ""  # a może tak?
    # return "en", ""
    # return "-", ""
    return "", ""


def getPremColor(n=control.setting("prem.identify")):
    if n == "0":
        return "blue"
    elif n == "1":
        return "red"
    elif n == "2":
        return "orange"
    elif n == "3":
        return "deeppink"
    elif n == "4":
        return "cyan"
    elif n == "5":
        return "lawngreen"
    elif n == "6":
        return "gold"
    elif n == "7":
        return "magenta"
    elif n == "8":
        return "yellowgreen"
    elif n == "9":
        return "nocolor"
    elif n == "10":
        return "none"
    else:
        return "nocolor"


def convert_size_to_bytes(size):
    suffixes = "","k","m","g","t"
    multipliers = {'{}b'.format(l) : 1024**i for i,l in enumerate(suffixes) }
    
    sre = re.compile("(\d+(?:[.,]\d+)?)\s?({})".format("|".join(x+"b" for x in suffixes)),re.IGNORECASE)

    def subfunc(m):
        return str(float(m.group(1).replace(',','.'))*multipliers[m.group(2).lower()])    
    
    bytes = sre.sub(subfunc,size)
    try:
        bytes = int(float( bytes ))
    except ValueError:
        bytes = '-'

    return bytes


def n2w(n):
    num2words = {0: 'zero', 1: 'one', 2: 'two', 3: 'three', 4: 'four',
                 5: 'five', 6: 'six', 7: 'seven', 8: 'eight', 9: 'nine',
                10: 'ten', 11: 'eleven', 12: 'twelve', 13: 'thirteen', 14: 'fourteen',
                15: 'fifteen', 16: 'sixteen', 17: 'seventeen', 18: 'eighteen', 19: 'nineteen',
                20: 'twenty', 30: 'thirty', 40: 'forty', 50: 'fifty',
                60: 'sixty', 70: 'seventy', 80: 'eighty', 90: 'ninety',
                }
    n = int(n)
    try:
        return num2words[n]
    except KeyError:
        try:
            return f'{num2words[n-n%10]} {num2words[n%10].lower()}'
        except KeyError:
            return n

# print( re.sub(r'\b\d{1,2}\b', lambda n: n2w(n.group()) , filename) )
